// kicad-pcb-value.cpp
//
// Robin Whittle  2016-02-23 rw@firstpr.com.au  http://www.firstpr.com.au/kicad/
//
// Changelog and version history below.
//
// Public domain, but if you want me to consider any additions or alterations
// then you must use my approach to fully documenting the code, including: 
// generally putting comments on non-code lines; use full sentences and
// proper punctuation; give proper description of variables and purposes of 
// the code; maintain the arrangements for comments normally starting at some
// column number (49 below) so the code is generally on the left and the
//  comments are generally on the right, and so are easilly distinguishable.
//
// In the future I intend to document this in a more formal manner as the
// "Dual Track" approach to programming, which can be applied to many 
// programming and probably hardware description (FPGA) languages.
//
//
// This program processes the netlist file generated by Eeschema, with the
// the modified version to be read into Pcbnew, to acheive these outcomes:
//
//    Normally, the contents of the "value" field in each component in the 
//    schematic is transferred to the contents of the "value" field of the 
//    corresponding footprint in the PCB design.  The contents of this 
//    "value" field in the schematic design is also used for generating the
//    the BOM.
//
//    This program helps with scenarios where we want a different piece of
//    text, or nothing showing at all, for the value in the PCB design, with
//    respect to the value of the matching schematic component.
//
//    In such cases, the schematic component is edited in Eeschema to have an
//    extra field whose name is: "PCBVal".  The contents of this field cannot 
//    be empty, but can contain a single space.
//
//    If the value of the "PCBVal" field has '%' as its first and perhaps 
//    only character, then there will be no change as described below.  
//    Therefore a reasonable approach is to have, as a default for all 
//    schematic components a "PCBVal" field with its contents set to "%" and 
//    visibility turned off.  This means that it is impossible to use this 
//    facility to generate a value for the PCB component whose contents start
//    with '%'. 
//
//    For those devices ("device" here means a physically and logically 
//    unique electronic component, as represented by a schematic component in
//    the schematic file and by its matching footprint in Pcbnew's board 
//    design file) where we want the contents of the "value" field in the PCB
//    design to differ from that in the schematic, we add this extra field, 
//    named "PCBVal" to the schematic component and set its value to whatever 
//    we want to see in the PCB design as the "value" field.  If we want the 
//    latter to appear empty - so as not to clutter the component legend 
//    layer with stuff we don't want on the board, or on the screen in gray 
//    (as occurs when the footprint's value contains printable characters and
//    has its visibility attribute set to "invisible") - then we set the 
//    value of the PCBVal field to a single space.  
//    
//    Scenario 1: 
//
//        I want the value in the schematic to be the full part number of the
//        device, such as "BZT52H-B5V6" which specifies a Zener diode with a 
//        particular voltage and power handling capacity, in a particular SMT 
//        package.  This is the full manufacturer's part number - exactly 
//        what I need in the BOM.  Sophisticated BOM generation typically
//        involves further fields for full part numbers, but in this 
//        scenario, the point is that we want different text in the value
//        fields of the schematic and the PCB design.
//
//        It is way too long for the PCB, where the text would take up about 
//        three times the area of the Zener diode itself.
//
//        For the PCB, "5V6" is all I want.  This is suitable for helping 
//        with service work and is important since Zener diodes in SMT 
//        packages typically do not have their voltage marked on them.
//
//        I can make this happen by editing the value field of the diode's 
//        footprint in Pcbnew, but this will be overwritten next time I 
//        create an updated netlist in Eeschema, and read this into Pcbnew.
//
//        So I create the new field named "PCBVal" in this diode's parameters
//        in Eeschema (right-click > Edit component > Edit) and set its value
//        to "5V6" (without these quotes).  Alternatively, if this field
//        already exists, with its value consisting of just "%" or "%" 
//        followed by one or more characters, then I alter it to be just
//        "5V6".
//
//        If there were no other fields than "PCBVal", then the netlist
//        produced by Eeschema would have lines starting with "(fields" and 
//        "(field (name PCBVal) " for this component:
//
//            (comp (ref DZ1)
//                  (value BZT52H-B5V6)
//                  (footprint 0RWFP:SOD-123)
//                  (fields
//                    (field (name PCBVal) 5V6))
//                  (libsource (lib 0RW) (part D-ZENER-SOD-123))
//                  (sheetpath (names /2-DFR/) (tstamps /57146362/))
//                  (tstamp 57179AA3)) 
//
//        This program processes such a netlist to change the "value" field,
//        so the result is:
//
//            (comp (ref DZ1)
//                  (value 5V6)
//                  (footprint 0RWFP:SOD-123)
//                  (fields
//                    (field (name PCBVal) 5V6))
//                  (libsource (lib 0RW) (part D-ZENER-SOD-123))
//                  (sheetpath (names /2-DFR/) (tstamps /57146362/))
//                  (tstamp 57179AA3)) 
//
//        Pcbnew seems to make no use of the fields lines, and the value field
//        of the diode's footprint will be set to "5V6". 
//
//        Here I have described the change as being at the level of an 
//        individual component in a schematic.  However, it is possible to do
//        the same thing in Eeschema's library editor, so every instance in 
//        the schematic of some kind of library component will have this extra 
//        field "PCBVal", with its value set to either "#" so no changes are
//        made to the value accepted by Pcbnew, or to some other string which
//        will be used to make a change as described above.
//
//        It is also possible to make the text contents of this PCBVal field 
//        (the value of this field) appear on the schematic with any chosen 
//        position, orientation and text size.  For me it appears in purple if 
//        I set to visible.
//
//    Scenario 2: 
//
//        In the schematic, I make the value field invisible for a particular 
//        instance of a component, or in the library representation of such a 
//        component.  This is fine and good.  For instance, I have a 
//        symbol (schematic library part AKA component) for a single wire
//        solder pad for direct soldering to the PCB.  I may gang these up on 
//        the PCB for solderning a few wires of a ribbon cable.
//
//        Each library component has a name in the library, determined by the 
//        contents of its value field.  This one's name is "RIBBON-WIRE".  I 
//        make this hidden ("T" icon on top toolbar > Value > Visibility:
//        Show (not ticked) so I don't have to see this on the schematic.
//
//        If Pcbnew reads the unmodified netlist, the result will be a piece 
//        of gray text "RIBBON-WIRE" visible in the PCB design, where this
//        "value" field has its Visibility attribute set to "Invisible".
//        This means it will not appear on the physical board, but it 
//        still clutters up my screen with gray text.  The only solution 
//        is to ensure that for these footprints in the PCB design, this
//        "value" field is set to either no characters, or to one or more
//        spaces.
//
//        This program acheives this, once the schematic component has a
//        "PCBVal" field whos contents (value) set to a single space.  
//        (It cannot be 0 characters.)  This is represented in the netlist 
//        as:
//
//                  (fields
//                    (field (name PCBVal) " ")) 
//
//        or, more generally, when there is no second parenthesis due to 
//        one or more fields after the PCBVal field:
//
//                  (fields
//                                     << 0 or more lines for other fields.
//                    (field (name PCBVal) " ")
//                                     << 1 or more lines for other fields.
//
//
//        This program transfers this to the "value" field:
//
//                  (value " ") 
//
//        of the output file - the modified netlist.  Once the modified 
//        netlist is read into Pcbnew, the matching footprint no longer 
//        has any annoying gray text.
//
//
// Usage
// -----
//
//    The program requires the name of the input and output file names, 
//    without the .net extension, for instance, for an input file xyz.net:
//
//        kicad-pcb-value xyz xyz-mod 
//
//    The program does not report on the actual changes.  These are best
//    checked with a diff program or Beyond Compare.  The program will 
//    stop with assertion errors if unexpected conditions occur.  It 
//    reports the number of component definitions found and the number
//    changed.
//
//
// Compilation
// -----------
//
// This code requires the compiler have any required flags set for C++11
// compatibility.  It should be fine with any modern C++ compiler.
//
// I wrote it with Codelite 9.1.5 under Windows 7 and compiled it with the
// TDM version of the mingw version of gcc 5.1.0. 
//
//
// Changelog and version history
// -----------------------------
//
//  2016-04-23  Initial version 1.0.  Works for me with Kicad 4.0.2-stable.
//
//  2016-04-24  Version 1.1.  Copes with the PCBVal string being in quotes
//              and with parentheses in that string.  
//
//              V1.0 looks for the left-most right parenthesis as the 
//              delimiter.  This  does not work if there is a right 
//              parenthesis in the string.  Eeschema puts the string in 
//              quotes if it has at least one space of if there is a left 
//              or right parenthesis.
// 
//              The new algorithm works from the right of the string, where
//              the last two characters should be "))" and the character
//              immediately before this is the last character of the string
//              we want.  So the new algorithm doesn't look for quotes, but
//              will not be tripped up by parentheses in the PCBVal string.
//
//  2016-04-28  The above algorithm relies on two parentheses, but this is 
//              only the case if the PCBVal field is the last field.  The 
//              new algorithm for finding the contents of the PCBVal field
//              finds the leftmost of the either one or two final 
//              parentheses.  
//
//              Changes are only made if the first character of the value 
//              of the PCBVal field is not '%'.


#include <cstdlib>          // For exit().
#include <string>
#include <fstream>          // For fstream.
#include <iostream>         // For cout.
#include <cassert>          // For assert.

#include <vector>           // For the various Containers we are using. 

using std::cout;
using std::cin;
using std::endl;

using std::string;
using std::fstream;
using std::vector;




int main(int argc, char **argv)
{
                                                // The input file is read into a vector of strings,
                                                // one string for each line, without a newline at the
                                                // end of each string.  We work on the data there and 
                                                // then write the modified contents to the output file.
    vector<string> netList;
    
    
    string infileName;
    string outfileName;
    
    fstream infile;
    fstream outfile;    
                                                // Number of component definitions found and value 
                                                // elements within them changed.
    int compsFound =  0;
    int valsChanged = 0;
    
                                                // Indexes into netList[] for the first and last lines
                                                // of each component definition.  We find these by 
                                                // looking for:
                                                //
                                                //     "    (comp (ref"
                                                //     "      (tstamp" 
    int lineCompStart = 0;
    int lineCompEnd =   0;
     
                                                // Similarly for the lines which contain the 
                                                // "value xxxx" line, the start of the fields section, 
                                                // and the line which refers to the PCBVal field.  Look 
                                                // for:
                                                //
                                                //     "      (value"     
                                                //     "      (fields"
                                                //     "        (field (name PCBVal)"
    int lineValue =        0;
    int lineFields =       0;
    int lineFieldsPCBVal = 0;
    
                                                // Check there are two command line parameters.
    if (argc != 3)
    {
        cout << "kicad-pcb-value  V1.2  Robin Whittle 2016-04-28" << endl;
        cout << "See the source code at http://www.firstpr.com.au/kicad/ for full information." << endl;
        cout << endl;
        cout << "Usage: The two parameters are the names of the input and output files, " << endl;
        cout << "       without the .net extension." << endl;
        cout << endl;
        cout << "       So, for xyz.net:"  << endl;
        cout << endl;
        cout << "            kicad-pcb-value xyz xyz-mod" << endl;
        return 0;
    }    

    infileName =  argv[1];
    outfileName = argv[2]; 
    
    infileName.append(".net");
    outfileName.append(".net");
    
                                                // Initialise the class fstream objects with their 
                                                // filenames by using their .open() member function.
                                                //
                                                // On error, complain and exit - by returning from the 
                                                // main() function with a non-zero value, by which the
                                                // program can signal to the operating system there was 
                                                // an error.  
                                                //
                                                // Normally I place multiple statements and braces on
                                                // separate lines but this is low-key, obvious, code 
                                                // which I want tucked out of the way.

    infile.open(infileName, std::ios::in);
    if (! infile)      {cout << "Can't read " << infileName << "!\n";  exit(1);}

    outfile.open(outfileName, std::ios::out);
    if (! outfile)     {cout << "Can't write " << outfileName << "!\n";  exit(1);}
    
    
                                                // Populate the vector of strings.
    string tempStr;
    while (std::getline(infile, tempStr))
    {
        netList.push_back(tempStr);
    }

                                                // Step through the lines looking for the lines 
                                                // mentioned above.  When we find the last line of each 
                                                // component, we may alter the "(value xxxx" line of 
                                                // that component if we found a field line for PCBVal.
 
    for (unsigned int iLine = 0; iLine < netList.size(); iLine++)
    {
        
        if      ((netList[iLine].compare(0, 14, "    (comp (ref"               ) ) == 0 )
        {
            lineCompStart    = iLine; 

                                                // This is the first match of a new component 
                                                // definition.  Zero the line numbers for the fields 
                                                // section, since we do not assume that all component 
                                                // definitions will have these.
            lineFields       = 0; 
            lineFieldsPCBVal = 0; 
        }
        
        else if ((netList[iLine].compare(0, 12, "      (value"                 ) )  == 0 )
        {
            lineValue        = iLine; 
        }    
        
        else if ((netList[iLine].compare(0, 13, "      (fields"                ) ) == 0 )
        {
            lineFields       = iLine; 
        }    

        else if ((netList[iLine].compare(0, 29, "        (field (name PCBVal) ") ) == 0 )
        {
            lineFieldsPCBVal = iLine; 
        }    

        else if ((netList[iLine].compare(0, 13, "      (tstamp"                ) ) == 0 )
        {
            lineCompEnd      = iLine; 
                    
                                                // This is where we do all the work.  First check that
                                                // the line indices are set to reasonable values.
                                                //
                                                // Set some minimum line number for the start of the 
                                                // component definition, since we expect there to be a 
                                                // bunch of stuff at the start of the file which does 
                                                // not contain these.
            assert (lineCompStart > 10);
            
                                                // As far as I know, the "(value " line will directly 
                                                // follow the first line of the component definition.
                                                // However, here we allow it to be on the following or a 
                                                // 1 or more lines after.
                                            
            assert (lineCompStart < lineValue);
            
                                                // Keep a count of component definitions found.
            compsFound++;

                                                // Expect the "(fields" line to follow the "(value " 
                                                // line with there probably being at least the 
                                                // "(footprint" line in between.
                                                //
                                                // However, this field section is optional, so only do 
                                                // the some checks if we found such a line.  
                                                //
                                                // We only change a "(value" line of these field lines 
                                                // are present, so the code for making these changes is
                                                // part of these if statements.
            
            if (lineFields > 0)
            {    
                assert (lineValue < lineFields);
            
            
                                                // Similarly, not every fields section will have a line 
                                                // for field PCBVal.
                if (lineFieldsPCBVal != 0)
                {
                                                // String to contain the data-carrying part of the line 
                                                // which specifies the value of the PCBVal field. 
                    string pcbVal;

                                                // String we will use to alter the value of the "value" 
                                                // field of the component.
                    string newVal;                    
                                                // Location within that string of the first right 
                                                // parenthesis.
                    unsigned int parentPos = 0;
                    
                                                // Expect the "(field (name PCBVal)" line to follow the 
                                                // "(fields" line with potentially one or more lines in
                                                // between.
            
                    assert ( lineFields < lineFieldsPCBVal);

                                                // Expect the last line, as far as I know, in the 
                                                // component definition to follow the 
                                                // "(field (name PCBVal)" line with probably multiple 
                                                // lines in between.
            
                    assert ( lineFieldsPCBVal < lineCompEnd) ;
                    
                                                // All the line number conditions look correct for a 
                                                // component definition where we will change the value 
                                                // field.
                                                //
                                                // Assert that the value of field PCBVal looks valid.  
                                                // It should be either a number of alphanumeric 
                                                // characters before the right parenthesis, or 
                                                // something quoted, such as the quoted single space: 
                                                // " ".  e.g.:
                                                //
                                                //               1         2         3         4
                                                //     01234567890123456789012345678901234567890 
                                                //    
                                                //             (field (name PCBVal) 5V6))
                                                //             (field (name PCBVal) " "))
                                                //             (field (name PCBVal) 104)
                                                // 
                                                // The double right parentheses are only found if this
                                                // is the last field.
                                                //
                                                // We have already checked this line to character 18.
                                                //
                                                // Get characters 29 and above into a string.
                                            
                    pcbVal = netList[lineFieldsPCBVal].substr(29);

                                                // The result should be, for instance:
                                                //
                                                //               1         2         3         4
                                                //     01234567890123456789012345678901234567890 
                                                //    
                                                //     5V6))
                                                //     " "))
                                                //     "Front R133 (AA)"))
                                                //     104) 
                                                //
                                                // Find the position of either the single right
                                                // parenthesis or the left one of two.
                                                //
                                                // Assert the last character is a right parenthesis.
                                                
                    parentPos  = pcbVal.size() - 1;
                    assert(pcbVal[parentPos] == ')');
                    
                                                // Is the character before this a right parenthesis too?
                    parentPos--;
                    
                    if (pcbVal[parentPos] == ')')
                    {    
                                                // Yes, so decrement it again.  
                        parentPos--;
                                                        
                    }
                                                // If there was only one right parenthesis, then 
                                                // parentPos was not decremented a second time.  Either
                                                // way, it now points to the final character of the 
                                                // value of the PCBVal field.
                                                //
                                                // Assert that this position is 1 or greater.  This 
                                                // should ensure that there is at least one character 
                                                // for this field.
                    assert(parentPos >= 1);
                    
                                                // We want to keep the leftmost of these one or two 
                                                // trailing parentheses.  Then newVal will be ready to 
                                                // use for the change to the "(value" line.
                                                //
                                                // We add 1 to get the position of the parenthesis we
                                                // want and add a second 1 to make the length encompass
                                                // this position.

                    newVal = pcbVal.substr(0, parentPos + 2);
                    
                                                // Only proceed with the change if the first character
                                                // of newVal is not "%".
                                                
                    if (newVal[0] != '%')
                    {
                                                // Overwrite the old "(value" line with, for instance:
                                                //               1         2         3         4
                                                //     01234567890123456789012345678901234567890 
                                                //           (value 5V6)  
                                                //           (value " ")
                                            
                        netList[lineValue] = "      (value " + newVal;
                    
                                                // Keep a count of changes.
                        valsChanged++;                        
                    }
                }
                
            }    
                                                // We don't really care about checking lineCompEnd if 
                                                // there was no field line for PCBVal.
        }    
    }    

        
                                                // All the changes have now been made.  Write the lines
                                                // to the output file and close it.
    for (auto line : netList)
    {
        outfile << line << endl;
    }    
        
    outfile.close();
        
                                                // Display counts.
                                            
    cout << "Number of component definitions found: " << compsFound  << endl;
    cout << "Number of component values changed:    " << valsChanged << endl;
        
    return 0;
}

